#ifndef ARRAYS
#define ARRAYS

#include <memory>
#include <iostream>
#include <cstdio>
#include <cstddef>
#include "mpi.h"

//#define MIN(X,Y) ((X) < (Y) ? : (X) : (Y)),
#define MIN(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a < _b ? _a : _b; })
//#define MAX(X,Y) ((X) > (Y) ? : (X) : (Y)),
#define MAX(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })

template<unsigned int N>
void printCoord(int coord[]) {
  printf("(");
  for (int i = 0; i < N; i++) {
    printf("%i, ", coord[i]);
  }
  printf(")");
}

template<unsigned int N>
void printCoord(unsigned int coord[]) {
  printf("(");
  for (int i = 0; i < N; i++) {
    printf("%i, ", coord[i]);
  }
  printf(")");
}

// TODO add reference counting

template <int N>
struct num { };

// predeclare classes
template <class V, unsigned int N>
class SubArray;
template <class V, unsigned int N>
class RootArray;
template <class IT, unsigned int N>
class IdxIterator;

template <unsigned int N>
struct ArrayBase {
protected:
  unsigned int dimLengths[N];
  int dimStarts[N];
  int dimEnds[N]; // starting points in parent array

  // init lengths, starts, ends
  void init(const unsigned int dimL[], const int dimS[]) {
    #ifdef DEBUG
    //printf("init arr:\n");
    #endif
    for (int i = 0; i < N; i++) {
      dimLengths[i] = dimL[i];
      dimStarts[i] = dimS[i];
      dimEnds[i] = dimS[i] + dimL[i] - 1;
      if (((int)dimLengths[i]) <= 0) {
        dimLengths[i] = 0;
        dimEnds[i] = dimStarts[i]-1;
      }
      #ifdef DEBUG
      //printf("dim %i: [%i..%i] len %i\n", i, dimStarts[i], dimEnds[i], dimLengths[i]);
      #endif
    }
  }

public:
  bool isEmpty() {
    for (int d = 0; d < N; d++) {
      if (((int)this->dimLengths[d]) <= 0) return true;
    }
    return false;
  }
 
  const int getDimStart(const int dim) const {
    return this->dimStarts[dim];
  }

  int*  getDimStarts() {
    return dimStarts;
  }

  int getDimEnd(int dim) {
    return this->dimEnds[dim];
  }

  int*  getDimEnds() {
    return dimEnds;
  }

  // size of each dim
  unsigned int getDimSize(int dim) {
    return this->dimLengths[dim];
  }

  unsigned int*  getDimSizes() {
    return dimLengths;
  }

  void printBounds() {
    printCoord<N>(dimStarts);
    printf(" to ");
    printCoord<N>(dimEnds);
  }

  const unsigned int numDims() const {
    return N;
  }

  const int boundsPackedSize() const {
    return N * 2;
  }

  void packBounds(int *buf) {
    if (this->isEmpty()) {
      for (int i = 0; i < N; i++) buf[i] = this->dimStarts[i];
      for (int i = N; i < N*2; i++) buf[i] = 0;
    } else {
      for (int i = 0; i < N; i++) buf[i] = this->dimStarts[i];
      for (int i = N; i < N*2; i++) buf[i] = this->dimLengths[i-N];
    }
  }

  void unpackStarts(int *buf, int arr[]) {
    for (int i = 0; i < N; i++) arr[i] = buf[i];
  }

  void unpackSizes(int *buf, unsigned int arr[]) {
    for (int i = N; i < N*2; i++) arr[i-N] = buf[i];
  }

  bool isSubsetOf(ArrayBase<N> *b) {
    // if its empty then its a subset
    if (this->isEmpty()) return true;

    for (int i = 0; i < N; i++) {
      if (this->dimStarts[i] < b->dimStarts[i]) return false;
      if (this->dimEnds[i] > b->dimEnds[i]) return false; 
    }
    return true;
  }

};

// iterator that also carries integer indices
template <class IT, unsigned int N>
struct IdxIterator {
private:
  int startIdx[N], endIdx[N];  
  int curIdx[N];
  IT iter;  // iterator over that array

  // inc and dec WITHOUT incrementing internal iterator!
  // inc
  template <int dim>
  inline void inc(num<dim>) {
    #ifdef DEBUG
    if (dim > N-1 || dim < 0) {
      std::cerr << "IdxIterator::inc(" << dim << ") dim out of bounds!\n";
      throw 1;
    }
    #endif

    curIdx[dim]++;
    if (curIdx[dim] > endIdx[dim]) {
      if (dim > 0) {
        curIdx[dim] = startIdx[dim];
        inc(num<dim-1>());
      }
    }
  }
  inline void inc(num<-1>) {
    std::cerr << "IdxIterator::inc(" << -1 << ") dim out of bounds!\n";
    throw 1;
  }

  // dec
  template <int dim>
  inline void dec(num<dim>) {
    #ifdef DEBUG
    if (dim > N-1 || dim < 0) {
      std::cerr << "IdxIterator::dec(" << dim << ") dim out of bounds!\n";
      throw 1;
    }
    #endif

    curIdx[dim]--;
    if (curIdx[dim] < startIdx[dim]) {
      if (dim > 0) {
        curIdx[dim] = endIdx[dim];
        dec(num<dim-1>());
      }
    }
  }
  inline void dec(num<-1>) {
    std::cerr << "IdxIterator::dec(" << -1 << ") dim out of bounds!\n";
    throw 1;
  }

public:

  // copy constructor
  IdxIterator(const IdxIterator<IT, N> &r): iter(r.iter) {
		for (int i = 0; i < N; i++) {
			curIdx[i] = r.curIdx[i];
			startIdx[i] = r.startIdx[i];
			endIdx[i] = r.endIdx[i];
		}
  }

  // constructor
  IdxIterator(IT it, int cur[], int start[], int end[]): iter(it) {
    for (int i = 0; i < N; i++) {
      curIdx[i] = cur[i];
      startIdx[i] = start[i];
      endIdx[i] = end[i];
    }
  }

  // returns current index
  inline int* getIndex() {
    return curIdx;
  }

  // returns current internal iterator
  // which can be dereferenced to get current
  // element / value
  inline IT getElement() {
    return iter;
  }

  // prefix increment
  inline IdxIterator<IT, N> operator++() {
    iter++;
    inc(num<N-1>());
    return *this;
  }

  // postfix increment
  inline IdxIterator<IT, N> operator++(int) {
    IdxIterator<IT, N> i = *this;
    iter++;
    inc(num<N-1>());
    return i;
  }

  // prefix decrement
  inline IdxIterator<IT, N> operator--() {
    iter--;
    dec(num<N-1>());
    return *this;
  }

  // postfix decrement
  inline IdxIterator<IT, N> operator--(int) {
    IdxIterator<IT, N> i = *this;
    iter--;
    dec(num<N-1>());
    return i;
  }

  // equality comparison
  bool operator==(const IdxIterator<IT, N>& b) const {
    #ifdef DEBUG
    // check that index bounds are the same
    for (int i = 0; i < N; i++) {
      if (startIdx[i] != b.startIdx[i] ||
          endIdx[i] != b.endIdx[i]) {
        std::cerr << "IdxIterator trying to compare idx iterators with different start and end indices!\n"
                  << "dim " << i << ": a.startIdx = " << startIdx[i] << ", b.startIdx = " << b.startIdx[i] << "\n"
                  << "    " << "a.endIdx = " << endIdx[i] << ", b.endIdx = " << b.endIdx[i] << "\n";
        throw 1;
      }
    }
    // check that if internal iterators are the same, then
    if (iter == b.iter) {
      for (int i = 0; i < N; i++) {
        if (curIdx[i] != b.curIdx[i]) {
          std::cerr << "IdxIterator with same underlying iterators have different current indices!\n";
          for (int dim = 0; dim < N; dim++) {
            std::cerr << "a.curIdx[" << dim << "] = " << curIdx[dim] << "; b.curIdx = " << b.curIdx[dim] << "\n";
          }
          throw 1;
        }
      }
    }
    #endif

    return (iter == b.iter);
  }

  bool operator!=(const IdxIterator<IT, N>& b) const {
    return !(*this == b);
  }

};

// special case 0-dimensional array, has 
// no index
template <class IT>
struct IdxIterator<IT, 0> {
  IT iter;  // iterator over that array

  // copy constructor
  IdxIterator(const IdxIterator<IT, 0> &r): iter(r.iter) {}

  // constructor
  IdxIterator(IT it, int cur[], int start[], int end[]): iter(it) {}

  // returns current index
  int* getIndex() {
    return NULL;
  }

  // returns current internal iterator
  // which can be dereferenced to get current
  // element / value
  IT getElement() {
    return iter;
  }

  // prefix increment
  IdxIterator<IT, 0> operator++() {
    iter++;
    return *this;
  }

  // postfix increment
  IdxIterator<IT, 0> operator++(int) {
    IdxIterator<IT, 0> i = *this;
    iter++;
    return i;
  }

  // prefix decrement
  IdxIterator<IT, 0> operator--() {
    iter--;
    return *this;
  }

  // postfix decrement
  IdxIterator<IT, 0> operator--(int) {
    IdxIterator<IT, 0> i = *this;
    iter--;
    return i;
  }

  // equality comparison
  bool operator==(const IdxIterator<IT, 0>& b) const {
    return (iter == b.iter);
  }

  bool operator!=(const IdxIterator<IT, 0>& b) const {
    return !(*this == b);
  }

};

template <class V, unsigned int N>
struct RootArray: public ArrayBase<N> {
public:
  // types
  typedef V* iterator;
  typedef const V* const_iterator;
  typedef IdxIterator<iterator, N> idx_iterator;
  typedef size_t size_type;
  typedef V value_type;
  typedef V& reference;
  typedef const V& const_reference;

protected:
  enum { ndims = N };
  //unsigned int dimLengths[N];
  unsigned int nElems[N+1];
  //int dimStarts[N];
  iterator limit; 
  MPI::Datatype arrType;
  const MPI::Datatype& elType;
  bool exBuffer; // using external buffer? i.e. don't deallocate.

  void allocate() {
    // allocate space
    if (elems != NULL) delete [] elems; // if reinit - deallocate old
    elems = new V[nElems[0]+1]; // first element
    limit = &elems[nElems[0]]; // just past the end
  }

  void initRootArr(const unsigned int dimL[], const int dimS[]) {
    // assign starts, lengths, ends
    this->init(dimL, dimS);

    // assign lengths
    nElems[N] = 1;
    for (int i = N-1; i >= 0; i--) {
      nElems[i] = nElems[i+1] * dimL[i];
    }
  }

public:
  iterator elems;

public:
  // constructors
  
  // create uinitialized array, if don't know sizes yet, need to call initArr later
  RootArray(const MPI::Datatype& type): elems(NULL), arrType(MPI::DATATYPE_NULL), elType(type), exBuffer(false) {
  }

  // create initialized array, allocate buffer on heap
  RootArray(const MPI::Datatype& type, const unsigned int dimL[], const int dimS[]): elems(NULL), arrType(MPI::DATATYPE_NULL), elType(type), exBuffer(false) {
    initArr(dimL, dimS);
  }

  // create initialized array, using buffer provided
  RootArray(const MPI::Datatype& type, const unsigned int dimL[], const int dimS[], V* buffer): elems(buffer), arrType(MPI::DATATYPE_NULL), elType(type), exBuffer(true) {
    initRootArr(dimL, dimS);
  }

  void initArr(const unsigned int dimL[], const int dimS[]) {
    // init sizes
    initRootArr(dimL, dimS);
    // allocate buffer on heap
    allocate();
  }

  void attachBuffer(V* buffer) {
    exBuffer = true;
    elems = buffer;
  }

  bool isInit() {
    return (elems != NULL);
  }

  // destructor
  ~RootArray() {
    // deallocate data type if it is allocated
    if (arrType != MPI::DATATYPE_NULL) {
      arrType.Free();
    }
    // deallocate elements
    if (elems != NULL && !exBuffer) { 
      delete [] elems;
    }
  }

  // accessor
  inline reference operator[](int idx[N]) {
    // get offset
    /*unsigned int offset = 0;
    for (int i = 0; i < N; i++) {
      int idx1 = idx[i] - this->dimStarts[i];

      #ifdef DEBUG
      // check if index is in bounds
      if (idx1 < 0 || idx1 >= this->dimLengths[i]) {
        std::cerr << "RootArray index " << i << " out of bounds: " << idx1 << "!";
        throw 1;
      }
      #endif

      offset += idx1 * nElems[i+1];
    }
    return elems[offset];*/
    return elems[getOffset(num<N-1>(), idx)];
  }

  template <int dim>
  inline unsigned int getOffset(num<dim>, int idx[N]) {
    const int idx1 = idx[dim] - this->dimStarts[dim];

    #ifdef DEBUG
    // check if index is in bounds
    if (idx1 < 0 || idx1 >= this->dimLengths[dim]) {
      std::cerr << "RootArray index " << dim << " out of bounds: " << idx1 << "!";
      throw 1;
    }
    #endif

    return (idx1 * nElems[dim+1]) + getOffset(num<dim-1>(), idx);
  } 

  inline unsigned int getOffset(num<-1>, int idx[N]) {
    return 0;
  }

  // iterates in storage order
  iterator begin() { return elems; }
  const_iterator begin() const { return elems; }

  iterator end() { return limit; }
  const_iterator end() const { return limit; }

  idx_iterator beginIndex() { 
    return idx_iterator(begin(), this->dimStarts, this->dimStarts, this->dimEnds);
  }
  idx_iterator endIndex() {
    // TODO IS IT SAFE TO PASS THIS CUR TO THE CONSTRUCTOR? YES 
    // because it makes a copy internally.
    int cur[N];
    for (int i = 0; i < N; i++) cur[i] = this->dimStarts[i];
    cur[0] = this->dimEnds[0]+1;
    return idx_iterator(end(), cur, this->dimStarts, this->dimEnds);
  } 

  // equality test
  bool operator==(const RootArray<V, N>& b) const {
    if (b.ndims != ndims) return false;
    for (int i = 0; i < N; i++) {
      if (this->dimLengths[i] != b.dimLengths[i]) return false;
      if (this->dimStarts[i] != b.dimStarts[i]) return false;
    }
    for (unsigned int i = 0; i < nElems[0]; i++) {
      if (elems[i] != b.elems[i]) return false;
    }
    return true;
  }

  bool operator!=(const RootArray<V,N>& b) const {
    return !(*this == b);
  }

  /*bool operator<(const nullStruct& b) const {
    return false;
  }*/

  const MPI::Datatype& getElementType() const {
    // check have an eltype
    if (elType == MPI::DATATYPE_NULL) {
      std::cerr << "RootArray::getType: element type is null! Has array been initialized with initArr?\n";
      throw 1;
    }
    return elType;
  }

  // getMPIType... CONTIGUOUS * nElems[0]
  MPI::Datatype getType() {
    // check have an eltype
    if (elType == MPI::DATATYPE_NULL) {
      std::cerr << "RootArray::getType: element type is null! Has array been initialized with initArr?\n";
      throw 1;
    }

    // create data type if it doesn't already exist
    if (arrType == MPI::DATATYPE_NULL) {
      arrType = elType.Create_contiguous(nElems[0]);
      arrType.Commit();
    }
    return arrType;
  }  

  // TODO reorder, uses any order iterator to create
  // a copy of the array, stored with the new local index order
  // so that it can be sent as a contiguous block, and read 
  // using the storage order iterator.

  template <unsigned int NDIM>
  void partition(const int dimsToPart[], const unsigned int numParts[],
                 RootArray<SubArray<V, N>, NDIM> *result) {
    SubArray<V, N> me(this, this->dimLengths, this->dimStarts);
    int blockCoord[NDIM];
    for (int i = 0; i < NDIM; i++) blockCoord[i] = 0;
    me.partition<NDIM>(0, dimsToPart, numParts, me, blockCoord, result);
  }

  // TODO add partition function that, when indices are
  // in the correct order, splits up the array into contiguous
  // chunks (RootArrays), rather than sub arrays, for faster
  // transmission

};

// an iterator that skips blocks
template <class V, unsigned int N>
struct SkipIterator {
  private:
    enum { ndims = N };
    V* ptr;
    ptrdiff_t displs[N], displs2[N];
    int counter[N];
    unsigned int blockLens[N];

public:
  // constructor
  SkipIterator(V* start, int ctrs[], unsigned int blockL[], ptrdiff_t displL[]) {
     ptr = start;
     for (int i = N-1; i >= 0; i--) {
       blockLens[i] = blockL[i];
       displs[i] = displL[i];
       displs2[i] = displs[i] - displs2[i];
       counter[i] = ctrs[i];
     }
  }

  V& operator*() {
    return *ptr;
  }

  V& operator->() {
    return *ptr;
  }

  // increment iterator
  template <int dim>
  inline void inc(num<dim>) {
    counter[dim]++;
    if (counter[dim] == blockLens[dim]) {
      if (dim > 0) {
        counter[dim] = 0;
        inc(num<dim-1>());
      } else {
        ptr += displs[0];
      }
    } else {
      ptr += displs[dim];
    }
  }

  inline void inc(num<-1>) {
    std::cerr << "SkipIterator inc dim -1!\n";
    throw 1;
  }

  // prefix increment
  inline SkipIterator<V, N> operator++() {
    inc(num<N-1>());
    return *this;
  }

  // postfix increment
  inline SkipIterator<V, N> operator++(int) {
    SkipIterator<V, N> i = *this;
    inc(num<N-1>());
    return i;
  }

  // TODO FIX!

  // skips the iterator position forward
  // by n elements
  /*inline void skip(int dim, int n) {
    counter[dim] += n;
    int numLens = counter[dim] / blockLens[dim];
    if (numLens > 0) {
      if (dim == 0) {
        // move to end
        for (int i = 1; i < N; i++) counter[i] = 0;
        counter[0] = blockLens[0];
        //std::cerr << "SkipIterator trying to skip by " << n 
        //          << ", in dim " << dim << " which puts us at " << counter[dim]
        //          << " when the blocklength for this dim is " << blockLens[dim] << "\n";
        //throw 1;        
      } else {
        int remainder = counter[dim] % blockLens[dim];
        //ptr += (numLens * blockLens[dim]);
        counter[dim] = remainder;
        skip(dim-1, numLens);
      }
    } else {
      ptr += displs[dim] * n;
    }
  }*/

  // decrement iterator
  template <int dim>
  inline void dec(num<dim>) {
    counter[dim]--;
    if (counter[dim] < 0) {
      if (dim > 0) {
        counter[dim] = blockLens[dim] - 1;
        dec(num<dim-1>());
      } else {
        ptr -= displs[dim];
      }
    } else {
      ptr -= displs[dim];
    }
  }

  inline void dec(num<-1>) {
    std::cerr << "SkipIterator dec dim -1!\n";
    throw 1;
  }

  // prefix decrement
  inline SkipIterator<V, N> operator--() {
    dec(N-1);
    return *this;
  }

  // postfix decrement
  inline SkipIterator<V, N> operator--(int) {
    SkipIterator<V, N> i = *this;
    dec(N-1);
    return i;
  }

  // equality comparison
  bool operator==(const SkipIterator& b) const {
    //printf("a[%i,%i] at %p; b[%i,%i] at %p\n", counter[0], counter[1], ptr, b.counter[0], b.counter[1], b.ptr);

    #ifdef DEBUG
    // check if comparing that block lens etc are same
    if (ndims != b.ndims) {
      std::cerr << "SkipIterators being compared with different number of dims! " << ndims << " != " << b.ndims;
      throw 1;
    }
    for (int i = 0; i < N; i++) {
      if (blockLens[i] != b.blockLens[i] ||
          displs[i] != b.displs[i]) {
        std::cerr << "SkipIterators being compared with different block or skip lens!\n"
                  << "a.blockLen[" << i << "] = " << blockLens[i] << "; b.blockLen = " << b.blockLens[i] << "\n"
                  << "a.displs[" << i << "] = " << displs[i] << "; b.displs = " << b.displs[i] << "\n";
        throw 1;
      }
    }
    // check if pointers are same, but counters are different
    if (ptr == b.ptr) {
      for (int i = 0; i < N; i++) {
        if (counter[i] != b.counter[i]) {
          std::cerr << "SkipIterator pointers equal, but counters not!\n";
          throw 1;
        }
      }
    }
    #endif
    return (ptr == b.ptr);
  }

  bool operator!=(const SkipIterator& b) const {
    return !(*this == b);
  }
    
};

// a sub array
template <class V, unsigned int N>
struct SubArray: public ArrayBase<N> {
public:
  typedef SkipIterator<V, N> iterator;
  typedef IdxIterator<iterator, N> idx_iterator;
  typedef V& reference;
  typedef V* pointer;

protected:
  enum { ndims = N };
public:
  RootArray<V, N> *parent;
protected:
  ptrdiff_t dimDispls[N];

  MPI::Datatype arrType; 
  int *blockLens, *displs;
  bool displsInit;

  // calculate displacements between blocks
  inline void initDispls() {
    if (!displsInit) {
      int b[N], e[N]; // begin and end coordinates
      for (int i = N-1; i >= 0; i--) {
        // if the length of this dim is 1,
        // then what do we do?
        // TODO ...

        // dims before current all start
        for (int dim = 0; dim < i; dim++) {
          b[dim] = this->dimStarts[dim];
          e[dim] = this->dimStarts[dim];
        }
        // current dim start and start+1
        b[i] = this->dimStarts[i];
        e[i] = this->dimStarts[i]+1;
        // dims after current are end to start
        for (int dim = i+1; dim < N; dim++) {
          b[dim] = this->dimEnds[dim];
          e[dim] = this->dimStarts[dim]; 
        }

        /*printf("displ between ");
        printCoord<N>(b);
        printf(" and ");
        printCoord<N>(e);
        printf("\n");*/

        // use coordinately to calculate displacement for 
        // this dimension
        dimDispls[i] = &((*parent)[e]) - &((*parent)[b]);
       
        #if DEBUG
        //printf("dim(%i): start=%i, length=%i, displ=%i\n", i, this->dimStarts[i], this->dimLengths[i], (int)(dimDispls[i]));
        //printf("%p to %p\n", &((*parent)[b]), &((*parent)[e]));
        //for (int dim = 0; dim < N; dim++) {
        //  printf("(%4i) to (%4i)\n", b[dim], e[dim]);
        //}
        //printf("\n");
        #endif
      }
      displsInit = true;
    }
  }

public:
  // default constructor
  SubArray(): arrType(MPI::DATATYPE_NULL), parent(NULL), blockLens(NULL), displs(NULL), displsInit(false) {
  }

  // initializer to call us used default constructor
  void initArr(RootArray<V, N> *root, const unsigned int dimL[], const int dimS[]) {
    // init
    this->parent = root;
    this->init(dimL, dimS);

    // check these indices are a subset of the root
    if (!this->isSubsetOf(root)) {
      std::cerr << "SubArray:initArr: these indices are not a subset of parent's indices:\n";
      for (int i = 0; i < N; i++) {
        std::cerr << "this[" << i << "] is from " << this->getDimStart(i) << " to " << this->getDimEnd(i) << "; "; 
        std::cerr << "root[" << i << "] is from " << root->getDimStart(i) << " to " << root->getDimEnd(i) << "\n";
      }
      throw 1;
    }
  }

  // initializer to call us used default constructor
  void initArr(SubArray<V, N> *parent, const unsigned int dimL[], const int dimS[]) {
    initArr(parent->parent, dimL, dimS);
  }

  void initArr(RootArray<V, N> *root) {
    initArr(root, root->getDimSizes(), root->getDimStarts());
  }

  void initArr(const MPI::Datatype &type, const unsigned int dimL[], const int dimS[]) {
    // TODO NEED SMART POINTER TO REF COUNT ROOT
    RootArray<V, N> *root = new RootArray<V, N>(type, dimL, dimS);
    initArr(root);
  }

  void initArr(const MPI::Datatype &type) {
    // TODO NEED SMART POINTER TO REF COUNT ROOT
    RootArray<V, N> *root = new RootArray<V, N>(type);
    this->parent = root;
  }

  SubArray(RootArray<V, N> *root): arrType(root->getElementType().Dup()), blockLens(NULL), displs(NULL), displsInit(false) {
    initArr(root);
  }

    // constructor from root array
  SubArray(RootArray<V, N> *root, const unsigned int dimL[], const int dimS[]): arrType(MPI::DATATYPE_NULL), blockLens(NULL), displs(NULL), displsInit(false) {
    initArr(root, dimL, dimS);
  }

  // constructor from another subarray
  SubArray(SubArray<V, N> *parent, const unsigned int dimL[], const int dimS[]): arrType(MPI::DATATYPE_NULL), blockLens(NULL), displs(NULL), displsInit(false)  {
    initArr(parent, dimL, dimS);
  }

  // constructor that creates a root array
  // TODO needs ref counting
  SubArray(const MPI::Datatype &type): 
    arrType(MPI::DATATYPE_NULL), parent(NULL), blockLens(NULL), displs(NULL), displsInit(false) 
  {
    initArr(type);
  }

  // constructor that creates a root array
  SubArray(const MPI::Datatype &type, const unsigned int dimL[], const int dimS[]): 
    arrType(MPI::DATATYPE_NULL), parent(NULL), blockLens(NULL), displs(NULL), displsInit(false) 
  {
    initArr(type, dimL, dimS);
  }

  bool isInit() {
    return parent != NULL;
  }
  
  // destructor
  ~SubArray() {
     if (arrType != MPI::DATATYPE_NULL) {
       arrType.Free();
     }
     if (blockLens != NULL) {
       delete [] blockLens;
     }
     if (displs != NULL) {
       delete [] displs;

     }
   }

  RootArray<V, N>* getParent() {
   return parent;
  }

  // accessor
  inline reference operator[](int idx[N]) {
    #ifdef DEBUG
    for (int i = 0; i < N; i++) {
      int idx1 = idx[i] - this->dimStarts[i];
      // check if index is in bounds
      if (idx1 < 0 || idx1 >= this->dimLengths[i]) {
        std::cerr << "SubArray index " << i << " out of bounds: " << idx[i] << " (arr offset " << idx1 << ")!";
        throw 1;
      }
    }
    #endif
    return (*parent)[idx];
  }

  // accessor
  bool operator==(const SubArray& b) const {
    if (b.ndims != ndims) return false;
    for (int i = 0; i < N; i++) {
      if (this->dimLengths[i] != b.dimLengths[i]) return false;
      if (this->dimStarts[i] != b.dimStarts[i]) return false;
    }
    for (iterator ita = begin(), itb = b.begin(); ita != end(); ita++, itb++) {
      if (*ita != *itb) return false;
    }
    return true;
  }

  // iterate in storage order
  iterator begin() {
    // init displs if not already init
    initDispls();
    // init counters
    int ctr[N];
    for (int i = 0; i < N; i++) ctr[i] = 0; 
    // calculate start offset
    V* start = &((*parent)[this->dimStarts]);
    // return
    SkipIterator<V, N> it(start, ctr, this->dimLengths, dimDispls);
    return it;
  }
  idx_iterator beginIndex() {
    #ifdef DEBUG
    printf("beginIdx: curIdx, start, end\n");
    for (int i = 0; i < N; i++) {
      printf("          (%3i), (%3i) to (%3i)\n", this->dimStarts[i], this->dimStarts[i], this->dimEnds[i]);
    }
    #endif
    return idx_iterator(begin(), this->dimStarts, this->dimStarts, this->dimEnds);
  }

  iterator end() {
    // init displs if not already init
    initDispls();
    // init counters
    int ctr[N];
    for (int i = 0; i < N; i++) ctr[i] = 0;
    ctr[0] = this->dimLengths[0]; // should be the element just after the last one
    // calculate end offset
    V* end = &((*parent)[this->dimEnds]); end += dimDispls[0];
    SkipIterator<V, N> it(end, ctr, this->dimLengths, dimDispls);
    return it;
  }
  idx_iterator endIndex() {
    int cur[N];
    for (int i = 0; i < N; i++) cur[i] = this->dimStarts[i];
    cur[0] = this->dimEnds[0]+1;
    return idx_iterator(end(), cur, this->dimStarts, this->dimEnds);
  }

  inline const pointer getStartPointer() {
    return &(*this)[this->getDimStarts()];
  }

  // returns MPI element type
  const MPI::Datatype& getElementType() {
    return parent->getElementType();
  }

  // returns MPI array type
  MPI::Datatype getType() {
    if (arrType == MPI::DATATYPE_NULL) {
      // multidimensional arrays
      if (N > 1) {
        // check if this array is same size and shape as root
        bool sameAsRoot = true;
        for (int i = 0; i < N; i++) sameAsRoot &= (this->dimStarts[i] == parent->getDimStart(i));
        for (int i = 0; i < N; i++) sameAsRoot &= (this->dimLengths[i] == parent->getDimSize(i));
        if (sameAsRoot) return parent->getType().Dup();

		    // calculate number of blocks and allocate arrays
		    int numBlocks = 1;
		    for (int i = 0; i < N-1; i++) numBlocks *= this->dimLengths[i];
		    displs = new int[numBlocks];
		    blockLens = new int[numBlocks];

		    // calculate block lengths
		    int blockLen = this->dimLengths[N-1];
		    for (int i = 0; i < numBlocks; i++) blockLens[i] = blockLen;

        #ifdef DEBUG
        //printf("numblocks: %i\n", numBlocks);
        //printf("blockLen: %i\n", blockLen);
        #endif

		    // calculate displacement from start of each block
		    /*iterator iter = begin();
        pointer startPtr = &(*iter);
		    for (int i = 0; i < numBlocks; i++) {
		      ptrdiff_t diff = &(*iter) - startPtr;
          displs[i] = (int)diff;
          #ifdef DEBUG
          printf("displ[%i] = %i\n", i, displs[i]);
          #endif
		      iter.skip(N-1, blockLen);
		    }*/
        // init index

        int idx[N]; 
        for (int i = 0; i < N; i++) idx[i] = this->dimStarts[i];
        // save displacements
        //printf ("parent starts pointer %p\n", parent->getDimStarts());
        //printf("parent starts dim 0 %i\n", parent->getDimStart(0));
        //printf("parent starts dim 1 %i\n", parent->getDimStart(1));
        pointer startPtr = &((*parent)[idx]);
        for (int i = 0; i < numBlocks; i++) {
          // save displ
          displs[i] = &((*parent)[idx]) - startPtr;
          //printf("displ[%i] = %i\n", i, displs[i]);
          // increment idx
          for (int dim = N-2; dim >= 0; dim--) {
            idx[dim]++;
            if (idx[dim] > this->dimEnds[dim]) idx[dim] = 0;
            else break;
          }
        }

		    // create type
		    arrType = parent->getElementType().Create_indexed(numBlocks, blockLens, displs);
      }
      // 1D arrays 
      else {
        arrType = parent->getElementType().Create_contiguous(this->dimLengths[0]);
      }
      arrType.Commit();
    }
    return arrType;
  }

  // partitions along one dimension 
  void partitionBlocked(const int dim, const int nparts, SubArray<V, N> *parts) {
    //printf("started partitionBlocked\n");
    #ifdef DEBUG
    if (dim < 0 || dim >= N) {
      std::cerr << "SubArray::partitionBlocked, invalid dim " << dim;
      throw 1;
    }
    #endif 

    // sizes of partitions
    int size = this->dimLengths[dim];
    int partBaseSize = size / nparts;
    int numExtras = size % nparts;

    for (int pIdx = 0; pIdx < nparts; pIdx++) {
      // start coord of partition
      int startCoord[N];
      for (int d = 0; d < N; d++)
        startCoord[d] = this->dimStarts[d];
      startCoord[dim] += (pIdx * partBaseSize) + MIN(numExtras, pIdx);

      // length partition
      unsigned int lenCoord[N];
      for (int d = 0; d < N; d++)
        lenCoord[d] = this->dimLengths[d];
      lenCoord[dim] = partBaseSize;
      if (pIdx < numExtras) lenCoord[dim]++;

      // add to array
      //printf("partitionBlocked: initArr(%i): ", pIdx);
      //printCoord<N>(lenCoord);
      //printf(" ");
      //printCoord<N>(startCoord);
      parts[pIdx].initArr(this, lenCoord, startCoord);

    }
  }

  void printDims() {
    printf("starts: ");
    printCoord<N>(this->dimStarts);
    printf(", lengths: ");
    printCoord<N>(this->dimLengths);
  }

};

  // partition along NDIM dimensions
  template <class V, unsigned int N, unsigned int NDIM>
  static void partitionArray(int dim, const int dimsToPart[], const unsigned int numParts[], 
                 SubArray<V, N> *block, int blockCoord[],
                 RootArray<SubArray<V, N>, NDIM> *result) {
   #ifdef DEBUG 
   printf("entered partitionArray(Recur): N=%u, NDIM=%u, dim=%i\n", N, NDIM, dim);
   for (int i = 0; i < NDIM; i++) {
     printf("  dimsToPart[%i] = %i, numParts[%i] = %u\n", i, dimsToPart[i], i, numParts[i]);
   }
   #endif

    if (dim == NDIM) {
      // base case (must copy block, as block will be deallocated)
      (*result)[blockCoord] = *block;

    } else {
      // recursive case, split block into n subarrays
      int nparts = numParts[dim];
      int arrDim = dimsToPart[dim];
      SubArray<V, N> *blocks = new SubArray<V,N>[nparts];
      //printf("blocks1: %p\n", blocks);
      block->partitionBlocked(arrDim, nparts, blocks);

      // iterate over new blocks
      blockCoord[dim] = 0;
      for (int i = 0; i < nparts; i++) {
        partitionArray<V, N, NDIM>(dim+1, dimsToPart, numParts, &blocks[i], blockCoord, result);
        blockCoord[dim]++;
      }
 
      // free subarrays
      //printf("blocks2: %p\n", blocks);
      std::cout.flush();
      delete [] blocks;
    }
  }

  // partition along NDIM dimensions
  template <class V, unsigned int N, unsigned int NDIM>
  void partitionArray(const int dimsToPart[], const unsigned int numParts[],  
                 SubArray<V, N> *array,
                 RootArray<SubArray<V, N>, NDIM> *result) {
   #ifdef DEBUG 
   printf("entered partitionArray(Top): N=%u, NDIM=%u\n", N, NDIM);
   for (int i = 0; i < NDIM; i++) {
     printf("  dimsToPart[%i] = %i, numParts[%i] = %u\n", i, dimsToPart[i], i, numParts[i]);
   }
   #endif

    int blockCoord[NDIM];
    for (int i = 0; i < NDIM; i++) blockCoord[i] = 0;
    partitionArray<V, N, NDIM>(0, dimsToPart, numParts, array, blockCoord, result);
  }

// iterator for iterating over a subset
// of an arrays indices
template<unsigned int N>
struct GroupIterator: public ArrayBase<N> {
private:
  int curIdx[N];
  int dimLocalToOrig[N];
  bool hasDimMap;

  // inc
  inline void inc(int dim) {
    #ifdef DEBUG
    if (dim > N-1 || dim < 0) {
      std::cerr << "GroupIterator::inc(" << dim << ") dim out of bounds!\n";
      throw 1;
    }
    #endif

    curIdx[dim]++;
    if (curIdx[dim] > this->dimEnds[dim]) {
      if (dim > 0) {
        curIdx[dim] = this->dimStarts[dim];
        inc(dim-1);
      }
    }
  }

  // dec
  inline void dec(int dim) {
    #ifdef DEBUG
    if (dim > N-1 || dim < 0) {
      std::cerr << "GroupIterator::dec(" << dim << ") dim out of bounds!\n";
      throw 1;
    }
    #endif

    curIdx[dim]--;
    if (curIdx[dim] < this->dimStarts[dim]) {
      if (dim > 0) {
        curIdx[dim] = this->dimEnds[dim];
        dec(dim-1);
      }
    }
  } 

public:
  GroupIterator(): hasDimMap(false) {}
 
  // constructor
  GroupIterator(int cur[], unsigned int lens[], int starts[]): hasDimMap(false) {
    this->init(lens, starts);
    for (int i = 0; i < N; i++) curIdx[i] = cur[i];
  }

  template <class V, unsigned int N2>
  void initIt(ArrayBase<N2> &arr, const int ndims, const int dimsToTraverse[], const int cur[]) {
    // copy starts and lens in order given by dimsToTraverse
    int starts[ndims];
    unsigned int lens[ndims];
    for (int dim = 0; dim < ndims; dim++) {
      int srcDim = dimsToTraverse[dim];
      dimLocalToOrig[dim] = srcDim;
      starts[dim] = arr.getDimStart(srcDim);
      lens[dim] = arr.getDimSize(srcDim);
      curIdx[dim] = cur[srcDim];
    }
    hasDimMap = true;
    this->init(lens, starts);
  }

  template <class V, unsigned int N2>
  void initStart(ArrayBase<N2> &arr, const int ndims, const int dimsToTraverse[]) {
    int cur[ndims];
    for (int dim = 0; dim < ndims; dim++) cur[dim] = arr.getDimStart(dimsToTraverse[dim]);
    initIt<V, N2>(arr, ndims, dimsToTraverse, cur);
  }

  template <class V, unsigned int N2>
  void initEnd(ArrayBase<N2> &arr, const int ndims, const int dimsToTraverse[]) {
    int cur[ndims];
    for (int dim = 0; dim < ndims; dim++) cur[dim] = arr.getDimStart(dimsToTraverse[dim]);
    cur[0] = arr.getDimEnd(dimsToTraverse[0]) + 1;
    initIt<V, N2>(arr, ndims, dimsToTraverse, cur);
  }

  // returns current index
  int* getIndex() {
    return curIdx;
  }

  int getIndex(int dim) {
    return curIdx[dim];
  }

  // inserts the indices this iterator iterates over
  // back at the correct positions into the index array 
  // given 
  void getOriginalIndex(int *indices) {
    #ifdef DEBUG
    if (!hasDimMap) {
      std::cerr << "GroupIterator::getOriginalIndex: can't call this when dimsToTraverse hasn't been initialized!";
      throw 1;
    }
    #endif

    for (int dim = 0; dim < N; dim++) {
      indices[dimLocalToOrig[dim]] = curIdx[dim]; 
    }
  }

  // prefix increment
  inline GroupIterator<N> operator++() {
    inc(N-1);
    return *this;
  }

  // postfix increment
  inline GroupIterator<N> operator++(int) {
    GroupIterator<N> i = *this;
    inc(N-1);
    return i;
  }

  // prefix decrement
  inline GroupIterator<N> operator--() {
    dec(N-1);
    return *this;
  }

  // postfix decrement
  inline GroupIterator<N> operator--(int) {
    GroupIterator<N> i = *this;
    dec(N-1);
    return i;
  }

  // equality comparison
  bool operator==(const GroupIterator<N>& b) const {
    #ifdef DEBUG
    // check that index bounds are the same
    for (int i = 0; i < N; i++) {
      if (this->dimStarts[i] != b.dimStarts[i] ||
          this->dimEnds[i] != b.dimEnds[i]) {
        std::cerr << "GroupIterator trying to compare group iterators with different start and end indices!\n"
                  << "dim " << i << ": a.startIdx = " << this->dimStarts[i] << ", b.startIdx = " << b.dimStarts[i] << "\n"
                  << "    " << "a.endIdx = " << this->dimEnds[i] << ", b.endIdx = " << b.dimEnds[i] << "\n";
        throw 1;
      }
    }
    #endif

    // check if current indices are the same
    for (int i = 0; i < N; i++) {
      if (curIdx[i] != b.curIdx[i]) return false;
    }
    return true;
  }

  bool operator!=(const GroupIterator<N>& b) const {
    return !(*this == b);
  }

};

#endif

